@@ -1,6 +1,7 @@
 #!/usr/bin/env python
-import sys, stat, time, os
-from bup import options, git, index, drecurse
+
+import sys, stat, time, os, errno, re
+from bup import metadata, options, git, index, drecurse, hlinkdb
 from bup.helpers import *
 from bup.hashsplit import GIT_MODE_TREE, GIT_MODE_FILE
 
@@ -48,11 +49,29 @@ def check_index(reader):
     log('check: passed.\n')
 
 
-def update_index(top, excluded_paths):
+def clear_index(indexfile):
+    indexfiles = [indexfile, indexfile + '.meta', indexfile + '.hlink']
+    for indexfile in indexfiles:
+        path = git.repo(indexfile)
+        try:
+            os.remove(path)
+            if opt.verbose:
+                log('clear: removed %s\n' % path)
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+
+
+def update_index(top, excluded_paths, exclude_rxs):
+    # tmax and start must be epoch nanoseconds.
+    tmax = (time.time() - 1) * 10**9
     ri = index.Reader(indexfile)
-    wi = index.Writer(indexfile)
+    msw = index.MetaStoreWriter(indexfile + '.meta')
+    wi = index.Writer(indexfile, msw, tmax)
     rig = IterHelper(ri.iter(name=top))
-    tstart = int(time.time())
+    tstart = int(time.time()) * 10**9
+
+    hlinks = hlinkdb.HLinkDB(indexfile + '.hlink')
 
     hashgen = None
     if opt.fake_valid:
@@ -63,7 +82,8 @@ def update_index(top, excluded_paths):
     bup_dir = os.path.abspath(git.repo())
     for (path,pst) in drecurse.recursive_dirlist([top], xdev=opt.xdev,
                                                  bup_dir=bup_dir,
-                                                 excluded_paths=excluded_paths):
+                                                 excluded_paths=excluded_paths,
+                                                 exclude_rxs=exclude_rxs):
         if opt.verbose>=2 or (opt.verbose==1 and stat.S_ISDIR(pst.st_mode)):
             sys.stdout.write('%s\n' % path)
             sys.stdout.flush()
@@ -75,10 +95,36 @@ def update_index(top, excluded_paths):
             if rig.cur.exists():
                 rig.cur.set_deleted()
                 rig.cur.repack()
+                if rig.cur.nlink > 1 and not stat.S_ISDIR(rig.cur.mode):
+                    hlinks.del_path(rig.cur.name)
             rig.next()
         if rig.cur and rig.cur.name == path:    # paths that already existed
-            if pst:
-                rig.cur.from_stat(pst, tstart)
+            try:
+                meta = metadata.from_path(path, statinfo=pst)
+            except (OSError, IOError), e:
+                add_error(e)
+                rig.next()
+                continue
+            if not stat.S_ISDIR(rig.cur.mode) and rig.cur.nlink > 1:
+                hlinks.del_path(rig.cur.name)
+            if not stat.S_ISDIR(pst.st_mode) and pst.st_nlink > 1:
+                hlinks.add_path(path, pst.st_dev, pst.st_ino)
+            # Clear these so they don't bloat the store -- they're
+            # already in the index (since they vary a lot and they're
+            # fixed length).  If you've noticed "tmax", you might
+            # wonder why it's OK to do this, since that code may
+            # adjust (mangle) the index mtime and ctime -- producing
+            # fake values which must not end up in a .bupm.  However,
+            # it looks like that shouldn't be possible:  (1) When
+            # "save" validates the index entry, it always reads the
+            # metadata from the filesytem. (2) Metadata is only
+            # read/used from the index if hashvalid is true. (3) index
+            # always invalidates "faked" entries, because "old != new"
+            # in from_stat().
+            meta.ctime = meta.mtime = meta.atime = 0
+            meta_ofs = msw.store(meta)
+            rig.cur.from_stat(pst, meta_ofs, tstart,
+                              check_device=opt.check_device)
             if not (rig.cur.flags & index.IX_HASHVALID):
                 if hashgen:
                     (rig.cur.gitmode, rig.cur.sha) = hashgen(path)
@@ -88,9 +134,22 @@ def update_index(top, excluded_paths):
             rig.cur.repack()
             rig.next()
         else:  # new paths
-            wi.add(path, pst, hashgen = hashgen)
+            try:
+                meta = metadata.from_path(path, statinfo=pst)
+            except (OSError, IOError), e:
+                add_error(e)
+                continue
+            # See same assignment to 0, above, for rationale.
+            meta.atime = meta.mtime = meta.ctime = 0
+            meta_ofs = msw.store(meta)
+            wi.add(path, pst, meta_ofs, hashgen = hashgen)
+            if not stat.S_ISDIR(pst.st_mode) and pst.st_nlink > 1:
+                hlinks.add_path(path, pst.st_dev, pst.st_ino)
+
     progress('Indexing: %d, done.\n' % total)
     
+    hlinks.prepare_save()
+
     if ri.exists():
         ri.save()
         wi.flush()
@@ -101,7 +160,7 @@ def update_index(top, excluded_paths):
                 check_index(ri)
                 log('check: before merging: newfile\n')
                 check_index(wr)
-            mi = index.Writer(indexfile)
+            mi = index.Writer(indexfile, msw, tmax)
 
             for e in index.merge(ri, wr):
                 # FIXME: shouldn't we remove deleted entries eventually?  When?
@@ -114,6 +173,9 @@ def update_index(top, excluded_paths):
     else:
         wi.close()
 
+    msw.close()
+    hlinks.commit_save()
+
 
 optspec = """
 bup index <-p|m|s|u> [options...] <filenames...>
@@ -124,26 +186,43 @@ m,modified print only added/deleted/modified files (implies -p)
 s,status   print each filename with a status char (A/M/D) (implies -p)
 u,update   recursively update the index entries for the given file/dir names (default if no mode is specified)
 check      carefully check index file integrity
+clear      clear the default index
  Options:
 H,hash     print the hash for each object next to its name
 l,long     print more information about each file
+no-check-device don't invalidate an entry if the containing device changes
 fake-valid mark all index entries as up-to-date even if they aren't
 fake-invalid mark all index entries as invalid
 f,indexfile=  the name of the index file (normally BUP_DIR/bupindex)
 exclude=   a path to exclude from the backup (can be used more than once)
 exclude-from= a file that contains exclude paths (can be used more than once)
+exclude-rx= skip paths that match the unanchored regular expression
 v,verbose  increase log output (can be used more than once)
 x,xdev,one-file-system  don't cross filesystem boundaries
 """
 o = options.Options(optspec)
 (opt, flags, extra) = o.parse(sys.argv[1:])
 
-if not (opt.modified or opt['print'] or opt.status or opt.update or opt.check):
+if not (opt.modified or \
+        opt['print'] or \
+        opt.status or \
+        opt.update or \
+        opt.check or \
+        opt.clear):
     opt.update = 1
 if (opt.fake_valid or opt.fake_invalid) and not opt.update:
     o.fatal('--fake-{in,}valid are meaningless without -u')
 if opt.fake_valid and opt.fake_invalid:
     o.fatal('--fake-valid is incompatible with --fake-invalid')
+if opt.clear and opt.indexfile:
+    o.fatal('cannot clear an external index (via -f)')
+
+# FIXME: remove this once we account for timestamp races, i.e. index;
+# touch new-file; index.  It's possible for this to happen quickly
+# enough that new-file ends up with the same timestamp as the first
+# index, and then bup will ignore it.
+tick_start = time.time()
+time.sleep(1 - (tick_start - int(tick_start)))
 
 git.check_repo_or_die()
 indexfile = opt.indexfile or git.repo('bupindex')
@@ -154,15 +233,19 @@ if opt.check:
     log('check: starting initial check.\n')
     check_index(index.Reader(indexfile))
 
-excluded_paths = drecurse.parse_excludes(flags)
+if opt.clear:
+    log('clear: clearing index.\n')
+    clear_index(indexfile)
 
+excluded_paths = parse_excludes(flags, o.fatal)
+exclude_rxs = parse_rx_excludes(flags, o.fatal)
 paths = index.reduce_paths(extra)
 
 if opt.update:
     if not extra:
         o.fatal('update mode (-u) requested but no paths given')
     for (rp,path) in paths:
-        update_index(rp, excluded_paths)
+        update_index(rp, excluded_paths, exclude_rxs)
 
 if opt['print'] or opt.status or opt.modified:
     for (name, ent) in index.Reader(indexfile).filter(extra or ['']):
