@@ -6,6 +6,12 @@ from bup.helpers import *
 suspended_w = None
 dumb_server_mode = False
 
+
+def do_help(conn, junk):
+    conn.write('Commands:\n    %s\n' % '\n    '.join(sorted(commands)))
+    conn.ok()
+
+
 def _set_mode():
     global dumb_server_mode
     dumb_server_mode = os.path.exists(git.repo('bup-dumb-server'))
@@ -13,22 +19,31 @@ def _set_mode():
            % (dumb_server_mode and 'dumb' or 'smart'))
 
 
+def _init_session(reinit_with_new_repopath=None):
+    if reinit_with_new_repopath is None and git.repodir:
+        return
+    git.check_repo_or_die(reinit_with_new_repopath)
+    # OK. we now know the path is a proper repository. Record this path in the
+    # environment so that subprocesses inherit it and know where to operate.
+    os.environ['BUP_DIR'] = git.repodir
+    debug1('bup server: bupdir is %r\n' % git.repodir)
+    _set_mode()
+
+
 def init_dir(conn, arg):
     git.init_repo(arg)
     debug1('bup server: bupdir initialized: %r\n' % git.repodir)
-    _set_mode()
+    _init_session(arg)
     conn.ok()
 
 
 def set_dir(conn, arg):
-    git.check_repo_or_die(arg)
-    debug1('bup server: bupdir is %r\n' % git.repodir)
-    _set_mode()
+    _init_session(arg)
     conn.ok()
 
     
 def list_indexes(conn, junk):
-    git.check_repo_or_die()
+    _init_session()
     suffix = ''
     if dumb_server_mode:
         suffix = ' load'
@@ -39,7 +54,7 @@ def list_indexes(conn, junk):
 
 
 def send_index(conn, name):
-    git.check_repo_or_die()
+    _init_session()
     assert(name.find('/') < 0)
     assert(name.endswith('.idx'))
     idx = git.open_idx(git.repo('objects/pack/%s' % name))
@@ -50,7 +65,7 @@ def send_index(conn, name):
 
 def receive_objects_v2(conn, junk):
     global suspended_w
-    git.check_repo_or_die()
+    _init_session()
     suggested = set()
     if suspended_w:
         w = suspended_w
@@ -89,22 +104,16 @@ def receive_objects_v2(conn, junk):
         #debug2('read %d bytes\n' % n)
         _check(w, n, len(buf), 'object read: expected %d bytes, got %d\n')
         if not dumb_server_mode:
-            oldpack = w.exists(shar)
+            oldpack = w.exists(shar, want_source=True)
             if oldpack:
-                if oldpack == True or oldpack.endswith('.midx'):
-                    # FIXME: we shouldn't really have to know about midx files
-                    # at this layer.  But exists() on a midx doesn't return the
-                    # packname (since it doesn't know)... probably we should
-                    # just fix that deficiency of midx files eventually,
-                    # although it'll make the files bigger.  This method is
-                    # certainly not very efficient.
-                    oldpack = w.objcache.packname_containing(shar)
-                    debug2('new suggestion: %r\n' % oldpack)
-                    w.objcache.refresh()
+                assert(not oldpack == True)
                 assert(oldpack.endswith('.idx'))
                 (dir,name) = os.path.split(oldpack)
                 if not (name in suggested):
-                    debug1("bup server: suggesting index %s\n" % name)
+                    debug1("bup server: suggesting index %s\n"
+                           % git.shorten_hash(name))
+                    debug1("bup server:   because of object %s\n"
+                           % shar.encode('hex'))
                     conn.write('index %s\n' % name)
                     suggested.add(name)
                 continue
@@ -120,14 +129,14 @@ def _check(w, expected, actual, msg):
 
 
 def read_ref(conn, refname):
-    git.check_repo_or_die()
+    _init_session()
     r = git.read_ref(refname)
     conn.write('%s\n' % (r or '').encode('hex'))
     conn.ok()
 
 
 def update_ref(conn, refname):
-    git.check_repo_or_die()
+    _init_session()
     newval = conn.readline().strip()
     oldval = conn.readline().strip()
     git.update_ref(refname, newval.decode('hex'), oldval.decode('hex'))
@@ -137,7 +146,7 @@ def update_ref(conn, refname):
 cat_pipe = None
 def cat(conn, id):
     global cat_pipe
-    git.check_repo_or_die()
+    _init_session()
     if not cat_pipe:
         cat_pipe = git.CatPipe()
     try:
@@ -165,6 +174,8 @@ if extra:
 debug2('bup server: reading from stdin.\n')
 
 commands = {
+    'quit': None,
+    'help': do_help,
     'init-dir': init_dir,
     'set-dir': set_dir,
     'list-indexes': list_indexes,
