@@ -1,10 +1,14 @@
 #!/usr/bin/env python
 import sys, stat, urllib, mimetypes, posixpath, time
-import tornado.httpserver
-import tornado.ioloop
-import tornado.web
 from bup import options, git, vfs
 from bup.helpers import *
+try:
+    import tornado.httpserver
+    import tornado.ioloop
+    import tornado.web
+except ImportError:
+    log('error: cannot find the python "tornado" module; please install it\n')
+    sys.exit(1)
 
 handle_ctrl_c()
 
@@ -60,6 +64,7 @@ def _compute_dir_contents(n, path, show_hidden=False):
             display = sub.name + '@'
         else:
             size = sub.size()
+            size = (opt.human_readable and format_filesize(size)) or size
 
         yield (display, link + url_append, size)
 
@@ -178,8 +183,9 @@ class BupRequestHandler(tornado.web.RequestHandler):
 optspec = """
 bup web [[hostname]:port]
 --
+human-readable    display human readable file sizes (i.e. 3.9K, 4.7M)
 """
-o = options.Options('bup web', optspec)
+o = options.Options(optspec)
 (opt, flags, extra) = o.parse(sys.argv[1:])
 
 if len(extra) > 1:
@@ -211,7 +217,12 @@ if __name__ == "__main__":
     http_server = tornado.httpserver.HTTPServer(application)
     http_server.listen(address[1], address=address[0])
 
-    print "Serving HTTP on %s:%d..." % http_server._socket.getsockname()
+    try:
+        sock = http_server._socket # tornado < 2.0
+    except AttributeError, e:
+        sock = http_server._sockets.values()[0]
+
+    print "Serving HTTP on %s:%d..." % sock.getsockname()
     loop = tornado.ioloop.IOLoop.instance()
     loop.start()
 
