@@ -50,8 +50,10 @@ def testencode():
 
 @wvtest
 def testpacks():
-    subprocess.call(['rm','-rf', 'pybuptest.tmp'])
-    git.init_repo('pybuptest.tmp')
+    os.environ['BUP_MAIN_EXE'] = bupmain = '../../../bup'
+    os.environ['BUP_DIR'] = bupdir = 'pybuptest.tmp'
+    subprocess.call(['rm','-rf', bupdir])
+    git.init_repo(bupdir)
     git.verbose = 1
 
     w = git.PackWriter()
@@ -147,15 +149,16 @@ def test_check_repo_or_die():
     git.check_repo_or_die()
     WVPASS('check_repo_or_die')  # if we reach this point the call above passed
 
-    mode = os.stat('pybuptest.tmp').st_mode
-    os.chmod('pybuptest.tmp', 0000)
+    os.rename('pybuptest.tmp/objects/pack', 'pybuptest.tmp/objects/pack.tmp')
+    open('pybuptest.tmp/objects/pack', 'w').close()
     try:
         git.check_repo_or_die()
     except SystemExit, e:
         WVPASSEQ(e.code, 14)
     else:
         WVFAIL()
-    os.chmod('pybuptest.tmp', mode)
+    os.unlink('pybuptest.tmp/objects/pack')
+    os.rename('pybuptest.tmp/objects/pack.tmp', 'pybuptest.tmp/objects/pack')
 
     try:
         git.check_repo_or_die('nonexistantbup.tmp')
