@@ -4,8 +4,9 @@ The vfs.py library makes it possible to expose contents from bup's repository
 and abstracts internal name mangling and storage from the exposition layer.
 """
 import os, re, stat, time
-from bup import git
+from bup import git, metadata
 from helpers import *
+from bup.hashsplit import GIT_MODE_TREE, GIT_MODE_FILE
 
 EMPTY_SHA='\0'*20
 
@@ -42,11 +43,11 @@ def _treeget(hash):
     it = cp().get(hash.encode('hex'))
     type = it.next()
     assert(type == 'tree')
-    return git.treeparse(''.join(it))
+    return git.tree_decode(''.join(it))
 
 
 def _tree_decode(hash):
-    tree = [(int(name,16),stat.S_ISDIR(int(mode,8)),sha)
+    tree = [(int(name,16),stat.S_ISDIR(mode),sha)
             for (mode,name,sha)
             in _treeget(hash)]
     assert(tree == list(sorted(tree)))
@@ -170,9 +171,18 @@ class Node:
         self.hash = hash
         self.ctime = self.mtime = self.atime = 0
         self._subs = None
+        self._metadata = None
+
+    def __repr__(self):
+        return "<%s object at X - name:%r hash:%s parent:%r>" \
+            % (self.__class__, self.name, self.hash.encode('hex'),
+               self.parent.name if self.parent else None)
 
     def __cmp__(a, b):
-        return cmp(a and a.name or None, b and b.name or None)
+        if a is b:
+            return 0
+        return (cmp(a and a.parent, b and b.parent) or
+                cmp(a and a.name, b and b.name))
 
     def __iter__(self):
         return iter(self.subs())
@@ -288,6 +298,16 @@ class Node:
         """Open the current node. It is an error to open a non-file node."""
         raise NotFile('%s is not a regular file' % self.name)
 
+    def _populate_metadata(self):
+        # Only Dirs contain .bupm files, so by default, do nothing.
+        pass
+
+    def metadata(self):
+        """Return this Node's Metadata() object, if any."""
+        if self.parent:
+            self.parent._populate_metadata()
+        return self._metadata
+
 
 class File(Node):
     """A normal file from bup's repository."""
@@ -374,6 +394,22 @@ class FakeSymlink(Symlink):
 
 class Dir(Node):
     """A directory stored inside of bup's repository."""
+
+    def __init__(self, *args):
+        Node.__init__(self, *args)
+        self._bupm = None
+
+    def _populate_metadata(self):
+        if not self._subs:
+            self._mksubs()
+        if not self._bupm:
+            return
+        meta_stream = self._bupm.open()
+        self._metadata = metadata.Metadata.read(meta_stream)
+        for sub in self:
+            if not stat.S_ISDIR(sub.mode):
+                sub._metadata = metadata.Metadata.read(meta_stream)
+
     def _mksubs(self):
         self._subs = {}
         it = cp().get(self.hash.encode('hex'))
@@ -383,12 +419,14 @@ class Dir(Node):
             it = cp().get(self.hash.encode('hex') + ':')
             type = it.next()
         assert(type == 'tree')
-        for (mode,mangled_name,sha) in git.treeparse(''.join(it)):
-            mode = int(mode, 8)
+        for (mode,mangled_name,sha) in git.tree_decode(''.join(it)):
+            if mangled_name == '.bupm':
+                self._bupm = File(self, mangled_name, mode, sha, git.BUP_NORMAL)
+                continue
             name = mangled_name
             (name,bupmode) = git.demangle_name(mangled_name)
             if bupmode == git.BUP_CHUNKED:
-                mode = 0100644
+                mode = GIT_MODE_FILE
             if stat.S_ISDIR(mode):
                 self._subs[name] = Dir(self, name, mode, sha)
             elif stat.S_ISLNK(mode):
@@ -396,6 +434,17 @@ class Dir(Node):
             else:
                 self._subs[name] = File(self, name, mode, sha, bupmode)
 
+    def metadata(self):
+        """Return this Dir's Metadata() object, if any."""
+        self._populate_metadata()
+        return self._metadata
+
+    def metadata_file(self):
+        """Return this Dir's .bupm File, if any."""
+        if not self._subs:
+            self._mksubs()
+        return self._bupm
+
 
 class CommitDir(Node):
     """A directory that contains all commits that are reachable by a ref.
@@ -408,7 +457,7 @@ class CommitDir(Node):
     the number of commits grows big.
     """
     def __init__(self, parent, name):
-        Node.__init__(self, parent, name, 040000, EMPTY_SHA)
+        Node.__init__(self, parent, name, GIT_MODE_TREE, EMPTY_SHA)
 
     def _mksubs(self):
         self._subs = {}
@@ -436,13 +485,13 @@ class CommitDir(Node):
 class CommitList(Node):
     """A list of commits with hashes that start with the current node's name."""
     def __init__(self, parent, name):
-        Node.__init__(self, parent, name, 040000, EMPTY_SHA)
+        Node.__init__(self, parent, name, GIT_MODE_TREE, EMPTY_SHA)
         self.commits = {}
 
     def _mksubs(self):
         self._subs = {}
         for (name, (hash, date)) in self.commits.items():
-            n1 = Dir(self, name, 040000, hash)
+            n1 = Dir(self, name, GIT_MODE_TREE, hash)
             n1.ctime = n1.mtime = date
             self._subs[name] = n1
 
@@ -450,7 +499,7 @@ class CommitList(Node):
 class TagDir(Node):
     """A directory that contains all tags in the repository."""
     def __init__(self, parent, name):
-        Node.__init__(self, parent, name, 040000, EMPTY_SHA)
+        Node.__init__(self, parent, name, GIT_MODE_TREE, EMPTY_SHA)
 
     def _mksubs(self):
         self._subs = {}
@@ -472,7 +521,7 @@ class BranchList(Node):
     /.commit/??/ . The symlink is named after the commit date.
     """
     def __init__(self, parent, name, hash):
-        Node.__init__(self, parent, name, 040000, hash)
+        Node.__init__(self, parent, name, GIT_MODE_TREE, hash)
 
     def _mksubs(self):
         self._subs = {}
@@ -480,6 +529,7 @@ class BranchList(Node):
         tags = git.tags()
 
         revs = list(git.rev_list(self.hash.encode('hex')))
+        latest = revs[0]
         for (date, commit) in revs:
             l = time.localtime(date)
             ls = time.strftime('%Y-%m-%d-%H%M%S', l)
@@ -494,14 +544,12 @@ class BranchList(Node):
                 t1.ctime = t1.mtime = date
                 self._subs[tag] = t1
 
-        latest = max(revs)
-        if latest:
-            (date, commit) = latest
-            commithex = commit.encode('hex')
-            target = '../.commit/%s/%s' % (commithex[:2], commithex[2:])
-            n1 = FakeSymlink(self, 'latest', target)
-            n1.ctime = n1.mtime = date
-            self._subs['latest'] = n1
+        (date, commit) = latest
+        commithex = commit.encode('hex')
+        target = '../.commit/%s/%s' % (commithex[:2], commithex[2:])
+        n1 = FakeSymlink(self, 'latest', target)
+        n1.ctime = n1.mtime = date
+        self._subs['latest'] = n1
 
 
 class RefList(Node):
@@ -514,7 +562,7 @@ class RefList(Node):
     that are reachable via a ref (e.g. a branch).  See CommitDir for details.
     """
     def __init__(self, parent):
-        Node.__init__(self, parent, '/', 040000, EMPTY_SHA)
+        Node.__init__(self, parent, '/', GIT_MODE_TREE, EMPTY_SHA)
 
     def _mksubs(self):
         self._subs = {}
