@@ -2,7 +2,7 @@
 import sys, os, subprocess, signal, getopt
 
 argv = sys.argv
-exe = argv[0]
+exe = os.path.realpath(argv[0])
 exepath = os.path.split(exe)[0] or '.'
 exeprefix = os.path.split(os.path.abspath(exepath))[0]
 
@@ -29,8 +29,8 @@ from bup.helpers import *
 # after running 'bup newliner', the tty_width() ioctl won't work anymore
 os.environ['WIDTH'] = str(tty_width())
 
-def usage():
-    log('Usage: bup [-?|--help] [-d BUP_DIR] [--debug] [--profile]'
+def usage(msg=""):
+    log('Usage: bup [-?|--help] [-d BUP_DIR] [--debug] [--profile] '
         '<command> [options...]\n\n')
     common = dict(
         ftp = 'Browse backup sets using an ftp-like client',
@@ -62,6 +62,8 @@ def usage():
     
     log("See 'bup help COMMAND' for more information on " +
         "a specific command.\n")
+    if msg:
+        log("\n%s\n" % msg)
     sys.exit(99)
 
 
@@ -73,11 +75,9 @@ try:
     optspec = ['help', 'version', 'debug', 'profile', 'bup-dir=']
     global_args, subcmd = getopt.getopt(argv[1:], '?VDd:', optspec)
 except getopt.GetoptError, ex:
-    log('error: ' + ex.msg + '\n')
-    usage()
+    usage('error: %s' % ex.msg)
 
 help_requested = None
-dest_dir = None
 do_profile = False
 
 for opt in global_args:
@@ -91,10 +91,13 @@ for opt in global_args:
     elif opt[0] in ['--profile']:
         do_profile = True
     elif opt[0] in ['-d', '--bup-dir']:
-        dest_dir = opt[1]
+        os.environ['BUP_DIR'] = opt[1]
     else:
-        log('error: unexpected option "%s"\n' % opt[0])
-        usage()
+        usage('error: unexpected option "%s"' % opt[0])
+
+# Make BUP_DIR absolute, so we aren't affected by chdir (i.e. save -C, etc.).
+if 'BUP_DIR' in os.environ:
+    os.environ['BUP_DIR'] = os.path.abspath(os.environ['BUP_DIR'])
 
 if len(subcmd) == 0:
     if help_requested:
@@ -112,10 +115,6 @@ subcmd_name = subcmd[0]
 if not subcmd_name:
     usage()
 
-subcmd_env = os.environ
-if dest_dir:
-    subcmd_env.update({"BUP_DIR" : dest_dir})
-
 def subpath(s):
     sp = os.path.join(exepath, 'bup-%s' % s)
     if not os.path.exists(sp):
@@ -124,8 +123,7 @@ def subpath(s):
 
 subcmd[0] = subpath(subcmd_name)
 if not os.path.exists(subcmd[0]):
-    log('error: unknown command "%s"\n' % subcmd_name)
-    usage()
+    usage('error: unknown command "%s"' % subcmd_name)
 
 already_fixed = atoi(os.environ.get('BUP_FORCE_TTY'))
 if subcmd_name in ['mux', 'ftp', 'help']:
@@ -153,33 +151,44 @@ else:
     outf = None
     errf = None
 
+ret = 95
+p = None
+forward_signals = True
 
-class SigException(Exception):
-    def __init__(self, signum):
-        self.signum = signum
-        Exception.__init__(self, 'signal %d received' % signum)
 def handler(signum, frame):
-    raise SigException(signum)
+    debug1('\nbup: signal %d received\n' % signum)
+    if not p or not forward_signals:
+        return
+    if signum != signal.SIGTSTP:
+        os.kill(p.pid, signum)
+    else: # SIGTSTP: stop the child, then ourselves.
+        os.kill(p.pid, signal.SIGSTOP)
+        signal.signal(signal.SIGTSTP, signal.SIG_DFL)
+        os.kill(os.getpid(), signal.SIGTSTP)
+        # Back from suspend -- reestablish the handler.
+        signal.signal(signal.SIGTSTP, handler)
+    ret = 94
 
 signal.signal(signal.SIGTERM, handler)
 signal.signal(signal.SIGINT, handler)
+signal.signal(signal.SIGTSTP, handler)
+signal.signal(signal.SIGCONT, handler)
 
-ret = 95
-p = None
 try:
     try:
         c = (do_profile and [sys.executable, '-m', 'cProfile'] or []) + subcmd
-        p = subprocess.Popen(c, stdout=outf, stderr=errf, preexec_fn=force_tty)
+        if not n and not outf and not errf:
+            # shortcut when no bup-newliner stuff is needed
+            os.execvp(c[0], c)
+        else:
+            p = subprocess.Popen(c, stdout=outf, stderr=errf,
+                                 preexec_fn=force_tty)
         while 1:
             # if we get a signal while waiting, we have to keep waiting, just
             # in case our child doesn't die.
-            try:
-                ret = p.wait()
-                break
-            except SigException, e:
-                log('\nbup: %s\n' % e)
-                os.kill(p.pid, e.signum)
-                ret = 94
+            ret = p.wait()
+            forward_signals = False
+            break
     except OSError, e:
         log('%s: %s\n' % (subcmd[0], e))
         ret = 98
